.code
include lz77cd.asm

FILL_ACTUAL_WINDOW MACRO
	local start_label, end_label, start_copying_label, label1
	
start_label:
	mov eax, read_buffer_begin_valid_ptr
	cmp eax, read_buffer_end_valid_ptr
	jnz start_copying_label
	
	;read buffer is empty
	mov read_buffer_begin_valid_ptr, offset read_buffer
	mov read_buffer_end_valid_ptr, offset read_buffer
	push offset helper
	push read_buffer_length
	push offset read_buffer
	call fill_buffer
	pop ebx
	pop ebx
	pop ebx
	
	;test return value in eax
	cmp eax, -1 ;error
	jnz label1
	mov error_text, offset text_error_reading
	jmp global_error
label1:
	test eax, eax ;all ok
	jnz end_label ;EOF

	;shift pointer to read_buffer end valid
	mov ebx, offset read_buffer
	add ebx, helper
	mov read_buffer_end_valid_ptr, ebx
	
start_copying_label:
	;now is something in read_buffer
	;isn't actual window full?
	mov ebx, actual_window_end_valid_ptr
	cmp ebx, buffer_search_window_ptr
	jz end_label
	
	;start copying
	mov eax, read_buffer_begin_valid_ptr
	;mov ebx, actual_window_end_valid_ptr ;in ebx is already actual_window_end_valid_ptr
	mov cl, byte ptr [eax]
	mov byte ptr [ebx], cl
	inc eax
	mov read_buffer_begin_valid_ptr, eax
	MOVE_FORWARD <ebx>, <1>
	mov actual_window_end_valid_ptr, ebx
	
	jmp start_label
	
end_label:
ENDM
; *****************************************************************

SEARCH_FOR_STRING MACRO
	;This values must be in registers when this macro ends
	;eax == 1 - sequence longer then 2 bytes found; eax == 0 - sequence not found
	;if (eax == 1)
	;	esi == # of bytes to go back
	;	edi == # of bytes - length of sequence
	;
	;Use of registers within this macro:
	;	eax
	;	ebx - actual position in actual window
	;	ecx - actual position in search window
	;	edx - actual found length
	;	esi - latest pointer after string in search window
	;	edi - latest length
	
;	LOCAL start_l, no_new_sequence, end_l, move_back, end_moving, are_equal
;	local test_bounds, setting_registers, count_bounds, end_macro
	
	;initialization
	mov ebx, buffer_actual_window_ptr
	mov ecx, buffer_search_window_ptr
	xor edx, edx
	xor esi, esi
	mov edi, 2 ;length 1 or 2 is not interesting
	
start_l:
	mov al, byte ptr [ebx]
	cmp al, byte ptr [ecx]
	jz are_equal
	
	; bytes are different
	cmp edi, edx
	jae no_new_sequence
	
	; found sequence is longer
	mov edi, edx
	mov esi, ecx
	
no_new_sequence:
	mov ebx, buffer_actual_window_ptr
	test edx, edx
	jnz move_back
		MOVE_FORWARD <ecx>, 1
		jmp end_moving
	move_back:
		dec edx
		test edx, edx
		jz end_moving
		MOVE_BACKWARD <ecx>, <edx>
		xor edx, edx
	end_moving:
	
	jmp test_bounds
	
	
are_equal:
	inc edx
	MOVE_FORWARD <ebx>, 1
	MOVE_FORWARD <ecx>, 1
	

test_bounds:
	cmp ebx, actual_window_end_valid_ptr
	jz end_l
	
	; if (edx == 0 && ecx == buffer_actual_window_ptr) goto end_l else goto start_l;
	test edx, edx
	jnz start_l
	cmp ecx, buffer_actual_window_ptr
	jnz start_l
	
end_l:
	;is last sequence longer?
	cmp edi, edx
	jae setting_registers
	mov edi, edx
	mov esi, ecx

setting_registers:
	xor eax, eax
	cmp edi, 2
	jbe end_macro
	inc eax
	MOVE_BACKWARD <esi>, <edi>
	
	;esi contains addres to first byte of string in buffer, we need count of bytes
	;before actual window
	cmp esi, buffer_actual_window_ptr
	jnb count_bounds
	mov ebx, buffer_actual_window_ptr
	sub ebx, esi
	mov esi, ebx
	jmp end_macro
count_bounds:
	mov ebx, buffer_end_ptr
	sub ebx, esi
	mov ecx, buffer_actual_window_ptr
	sub ecx, buffer_start_ptr
	mov esi, ebx
	add esi, ecx
end_macro:

ENDM
; *****************************************************************

WRITE_FLAG MACRO
	;do not change eax, esi, edi!!!
	;flag is in eax
	;local label1
	
	clc
	test eax, eax
	jz label1
	stc
label1:
	;carry is set to our flag
	
	mov edx, flags
	rcl edx, 1
	mov flags, edx

ENDM
; *****************************************************************

WRITE_BYTE MACRO
	;copying byte
	mov eax, buffer_actual_window_ptr
	mov cl, byte ptr [eax]
	mov ebx, write_buffer_end_valid_ptr
	mov byte ptr [ebx], cl
	
	;write buffer pointers
	inc ebx
	mov write_buffer_end_valid_ptr, ebx
	
	;DO NOT MOVE window pointers!!! will be moved in main loop
	;MOVE_FORWARD <eax>, 1
	;mov buffer_actual_window_ptr, eax
	;mov eax, buffer_search_window_ptr
	;MOVE_FORWARD <eax>, 1
	;mov buffer_search_window_ptr, eax
ENDM
; *****************************************************************

WRITE_SEQUENCE MACRO
	;	esi == # of bytes to go back ;lower 11 bits
	;	edi == # of bytes - length of sequence; lower 5 bits
	
	mov ecx, esi
	mov ebx, edi
	rol bl, 3
	REPT 5
		shl bl, 1
		rcl cx, 1
	ENDM
	
	xchg cl, ch ;write BIG ENDIAN style
	
	;cx contains word to write
	mov eax, write_buffer_end_valid_ptr
	mov word ptr [eax], cx
	inc eax
	inc eax
	mov write_buffer_end_valid_ptr, eax
	
	;DO NOT MOVE window pointers!!! will be moved in main loop
	;mov eax, buffer_actual_window_ptr
	;MOVE_FORWARD <eax>, <edi>
	;mov buffer_actual_window_ptr, eax
	;mov eax, buffer_search_window_ptr
	;MOVE_FORWARD <eax>, <edi>
	;mov buffer_search_window_ptr, eax
ENDM
; *****************************************************************

SAVE_FLAGS_AND_WRITE_BUFFER MACRO
	local end
	
	;is something to save in write_buffer?
	mov eax, write_buffer_end_valid_ptr
	cmp eax, offset write_buffer
	jz end
	
	;shift flags to the left border of byte
	mov ecx, 8
	sub ecx, parts_written
	mov edx, flags
	test ecx, ecx
	shl edx, cl
	;save flags
	mov flags_buffer, dl
	push offset helper
	push 1
	push offset flags_buffer
	call save_buffer
	pop eax
	pop eax
	pop eax
	
	;save write_buffer
	push offset helper
	mov eax, write_buffer_end_valid_ptr
	sub eax, offset write_buffer
	push eax
	push offset write_buffer
	call save_buffer
	pop eax
	pop eax
	pop eax
end:

ENDM
; *****************************************************************

; *********************** COMPRESS FUNCTION *****************************
;int lz77_compress(int (* fill_buffer)(char *, int, int *), int (* save_buffer)(char *, int, int *))
public lz77_compress
lz77_compress proc fill_buffer:dword, save_buffer:dword
	local actual_window_end_valid_ptr:dword
	local read_buffer_begin_valid_ptr:dword, read_buffer_end_valid_ptr:dword
	local write_buffer_end_valid_ptr:dword
	local parts_written:dword
	local flags:dword ;bit flags LETTER = 0, BACK REFERENCE = 1
	
	;local variables initialization
	mov eax, buffer_actual_window_ptr
	mov actual_window_end_valid_ptr, eax
	mov ebx, offset read_buffer
	add ebx, read_buffer_length
	mov read_buffer_begin_valid_ptr, ebx
	mov read_buffer_end_valid_ptr, ebx
	mov write_buffer_end_valid_ptr, offset write_buffer
	xor eax, eax
	mov parts_written, eax
	mov flags, eax

; |---------------------------- buffer ----------------------------------------------------|
; |-actual window(32)-|------------------- search window (2080)----------------------------|
;
; ^ actual_window_end_valid_ptr
; ^ buffer_start_ptr
; ^ buffer_actual_window_ptr
;                     ^ buffer_search_window_ptr
;                                                                            buffer_end_ptr  ^ (points BEHIND buffer)
; |----------- read buffer --------|
;                                  ^ read_buffer_begin_valid_ptr
;                                  ^ read_buffer_end_valid_ptr
; |----------- write buffer -------|
; ^ write_buffer_end_valid_ptr
; *********************************************************************
	;Main loop
start_loop:
	FILL_ACTUAL_WINDOW
	mov eax, buffer_actual_window_ptr
	cmp eax, actual_window_end_valid_ptr
	jz end_loop ;empty actual window
	
	SEARCH_FOR_STRING ;main macro
	;eax == 1 - sequence longer then 2 bytes found; eax == 0 - sequence not found
	;if (eax == 1)
	;	esi == # of bytes to go back
	;	edi == # of bytes - length of sequence
	
	WRITE_FLAG ;doesn't change eax, esi, edi
	test eax, eax
	jnz write_sequence_label
	WRITE_BYTE
	mov edi, 1 ;to shift actual window appropriately
	jmp end_write_sequence
write_sequence_label:
	WRITE_SEQUENCE ; reads esi, edi
end_write_sequence:
	mov eax, parts_written
	inc eax
	mov parts_written, eax
	
	;shift actual window
	mov eax, buffer_actual_window_ptr
	MOVE_FORWARD <eax>, <edi>
	mov buffer_actual_window_ptr, eax
	mov eax, buffer_search_window_ptr
	MOVE_FORWARD <eax>, <edi>
	mov buffer_search_window_ptr, eax
	
	;save if 8 parts written
	cmp parts_written, 8
	jnz end_write_parts
	SAVE_FLAGS_AND_WRITE_BUFFER
	mov write_buffer_end_valid_ptr, offset write_buffer
	mov parts_written, 0
	
end_write_parts:

	jmp start_loop
end_loop:
	;save what remains in write buffer
	SAVE_FLAGS_AND_WRITE_BUFFER
	;is not necessary to move write_buffer_end_valid_ptr
	
; *********************************************************************
	
	xor eax, eax
        ret

global_error:
	cmp error_text, 0
	jz end_error
	push error_text
	mov eax, offset puts
	call eax
	pop eax
end_error:
	mov eax, 1
	ret

lz77_compress endp
; *********************** END OF COMPRESS FUNCTION *****************************
