.code
include lz77cd.asm

; *************************************************************************

FILL_READ_BUFFER MACRO ;leave eax == 0 if all is ok, eax == 1 if EOF
	; DO NOT MODIFY esi, edi
	local start_label, start_copying, set_pointers, start_filling, label1, end_label

start_label:

	;move contens of buffer to the left
		mov eax, read_buffer_begin_valid_ptr
		mov ebx, offset read_buffer
		mov edx, read_buffer_end_valid_ptr
		cmp eax, edx ;points eax to read_buffer_end_valid_ptr??
		jz set_pointers
	start_copying:
		mov cl, byte ptr [eax]
		mov byte ptr [ebx], cl
		inc ebx
		inc eax
		cmp eax, edx ;points eax to read_buffer_end_valid_ptr??
		jnz start_copying
	set_pointers:
		mov read_buffer_end_valid_ptr, ebx
		mov read_buffer_begin_valid_ptr, offset read_buffer
		
start_filling:
	mov eax, read_buffer_length
	mov ebx, read_buffer_end_valid_ptr
	sub ebx, offset read_buffer
	sub eax, ebx
	push offset helper
	push eax
	mov eax, read_buffer_end_valid_ptr
	push eax
	call fill_buffer
	pop ebx
	pop ebx
	pop ebx
	
	;test return value in eax
	cmp eax, -1 ;error
	jnz label7
	mov error_text, offset text_error_reading
	jmp global_error
label7:
	test eax, eax ;all ok
	jnz end_label ;EOF

	;shift pointer to read_buffer end valid
	mov ebx, read_buffer_end_valid_ptr
	add ebx, helper
	mov read_buffer_end_valid_ptr, ebx

end_label:

ENDM
; *************************************************************************

SAVE_WRITE_BUFFER MACRO

	push offset helper
	mov eax, write_buffer_end_valid_ptr
	sub eax, offset write_buffer
	push eax
	push offset write_buffer
	call save_buffer
	pop eax
	pop eax
	pop eax
	
ENDM
; *************************************************************************

PROCESS_SEQUENCE MACRO
	;we have guarantee that in read_buffer is at least two bytes
	local start_loop, end_loop, label1
	
	mov eax, read_buffer_begin_valid_ptr
	xor ebx, ebx
	mov bx, word ptr [eax]
	inc eax
	inc eax
	mov read_buffer_begin_valid_ptr, eax
	xchg bl, bh
	
	xor ecx, ecx
	REPT 5
		shr bx, 1
		rcr cl, 1
	ENDM
	shr cl, 3
	
	;now ebx contains #of bytes to go to the left (zero for 2048 bytes)
	;ecx contains # of bytes to copy (zero for 32 bytes)
	test ebx, ebx
	jnz label1
	mov ebx, 2048
label1:
	mov eax, buffer_current_position_ptr
	mov esi, eax
	MOVE_BACKWARD <eax>, <ebx>
	mov ebx, write_buffer_end_valid_ptr
	
	;if ecx contains 0, fill it with 32
	test ecx, ecx
	jnz start_loop
	mov ecx, 32
	
start_loop:
	test ecx, ecx
	jz end_loop
	mov dl, byte ptr [eax]
	mov byte ptr [esi], dl ;to buffer
	mov byte ptr [ebx], dl ;to write_buffer
	MOVE_FORWARD <eax>, 1
	MOVE_FORWARD <esi>, 1
	inc ebx
	dec ecx
	jmp start_loop
end_loop:
	mov buffer_current_position_ptr, esi
	mov write_buffer_end_valid_ptr, ebx
		
ENDM
; *********************** DECOMPRESS FUNCTION *****************************
;int lz77_decompress(int (* fill_buffer)(char *, int, int *), int (* save_buffer)(char *, int, int *))
public lz77_decompress
lz77_decompress proc fill_buffer:dword, save_buffer:dword
	local buffer_current_position_ptr:dword
	local read_buffer_begin_valid_ptr:dword, read_buffer_end_valid_ptr:dword
	local write_buffer_end_valid_ptr:dword
	local parts_left:dword
	local flags:dword ;bit flags LETTER = 0, BACK REFERENCE = 1

	;local variables initialization
	mov buffer_current_position_ptr, offset buffer
	mov ebx, offset read_buffer
	add ebx, read_buffer_length
	mov read_buffer_begin_valid_ptr, ebx
	mov read_buffer_end_valid_ptr, ebx
	mov write_buffer_end_valid_ptr, offset write_buffer
	xor eax, eax
	mov parts_left, eax
	mov flags, eax
	
	xor edi, edi ; edi = 0, needed for empty file
start_loop:
	;edi contains next flag or 2 if flag is unavailable
	mov esi, read_buffer_end_valid_ptr
	sub esi, read_buffer_begin_valid_ptr
	cmp esi, 1
	jnbe label1
	FILL_READ_BUFFER
	test eax, eax
	jz label1
	;end of input file
	test esi, esi
	jz end_loop
	;last byte is in read_buffer, test if file isn't corrupted
	test edi, edi
	jz label1
	;at the end should be 2 bytes and is only 1 (edi == 1) or file ends with flags (edi == 2)
	mov error_text, offset text_input_file_corrupted
	jmp global_error
	
label1:
	mov eax, parts_left
	test eax, eax
	jnz label2
	
	;read flags
	mov eax, read_buffer_begin_valid_ptr
	mov bl, byte ptr [eax]
	mov edi, 0
	shl bl, 1
	jnc label3
	inc edi
label3:
	mov flags, ebx
	inc eax
	mov read_buffer_begin_valid_ptr, eax
	mov parts_left, 8
	jmp start_loop
	
label2:
	;parts_left is greather than 0
	;edi contains flag
	test edi, edi
	jnz sequence_label
	
	;copy byte to output
	mov eax, read_buffer_begin_valid_ptr
	mov bl, byte ptr [eax]
	inc eax
	mov read_buffer_begin_valid_ptr, eax
	mov eax, write_buffer_end_valid_ptr
	mov byte ptr [eax], bl
	inc eax
	mov write_buffer_end_valid_ptr, eax
	mov eax, buffer_current_position_ptr
	mov byte ptr [eax], bl
	MOVE_FORWARD <eax>, 1
	mov buffer_current_position_ptr, eax

	jmp end_sequence_label
	
sequence_label:
	PROCESS_SEQUENCE
	
end_sequence_label:

	mov eax, parts_left
	dec eax
	mov parts_left, eax
	
	; read next flag
	mov edi, 2
	mov ebx, flags
	test eax, eax
	jz label4 ;next flag is unavailable
	mov edi, 0
	shl bl, 1
	jnc label4
	inc edi
label4:
	mov flags, ebx	

	SAVE_WRITE_BUFFER
	mov write_buffer_end_valid_ptr, offset write_buffer
	
	jmp start_loop

end_loop:
	SAVE_WRITE_BUFFER

; *********************************************************************

	xor eax, eax
	ret
	
global_error:
	cmp error_text, 0
	jz end_error
	push error_text
	mov eax, offset puts
	call eax
	pop eax
end_error:
	mov eax, 1
	ret

lz77_decompress endp
; *********************** END OF DECOMPRESS FUNCTION *****************************
